#!/usr/bin/env python2
#
#  Valosoitin - Audio player for EFL
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import logging

import dbus
import dbus.service

class ValosoitinDBus(dbus.service.Object):
    def __init__(self, parent):
        self.parent = parent
        dbus.service.Object.__init__(self, dbus.SessionBus(), "/org/mpris/MediaPlayer2", "org.mpris.MediaPlayer2")

        self.props = {
            "CanQuit":              True,
            "FullScreen":           False,
            "CanSetFullscreen":     False,
            "CanRaise":             True,
            "HasTrackList":         False,
            "Identity":             "Valosoitin audio player",
            "DesktopEntry":         "valosoitin",
            "SupportedUriSchemes":  ("file",),
            "SupportedMimeTypes":   ("audio/mpeg", "application/ogg"),
        }

        self.playerprops = {
            "PlaybackStatus": "Stopped",
            "LoopStatus": "None",
            "Rate": 1.0,
            "Shuffle": self.parent.playmodes.random.state,
            "Metadata": dbus.Dictionary({"mpris:trackid": "/org/mpris/MediaPlayer2/TrackList/NoTrack",}, signature="sv", variant_level=1),
            "Volume": self.parent.player.audio_volume_get(),
            "Position": 0, # Time in usecs
            "MinimumRate": 1.0,
            "MaximumRate": 1.0,
            "CanGoNext": True,
            "CanGoPrevious": True,
            "CanPlay": True,
            "CanPause": True,
            "CanSeek": True,
            "CanControl": True,
        }

        if self.parent.playmodes.single.state and self.parent.playmodes.repeat.state:
            self.playerprops["LoopStatus"] = "Track"
        elif self.parent.playmodes.repeat.state:
            self.playerprops["LoopStatus"] = "Playlist"

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='ss', out_signature='v')
    def Get(self, interface, prop):
        if interface == "org.mpris.MediaPlayer2":
            return self.props[prop]
        elif interface == "org.mpris.MediaPlayer2.Player":
            return self.playerprops[prop]

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='ssv')
    def Set(self, interface, prop, value):
        if interface == "org.mpris.MediaPlayer2":
            self.props[prop] = value
        elif interface == "org.mpris.MediaPlayer2.Player":
            self.playerprops[prop] = value
            if prop == "Shuffle":
                self.parent.playmodes.random.state = value
            elif prop == "LoopStatus":
                modes = self.parent.playmodes
                if value == "None":
                    modes.single.state = False
                    modes.repeat.state = False
                elif value == "Track":
                    modes.single.state = True
                    modes.repeat.state = True
                elif value == "Playlist":
                    modes.single.state = False
                    modes.repeat.state = True
                else:
                    logging.error("Invalid value")

    @dbus.service.method(dbus.PROPERTIES_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, interface):
        if interface == "org.mpris.MediaPlayer2":
            return self.props
        elif interface == "org.mpris.MediaPlayer2.Player":
            return self.playerprops


    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2',
                         in_signature='', out_signature='')
    def Raise(self):
        self.parent.raise_window()

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2',
                         in_signature='', out_signature='')
    def Quit(self):
        self.parent.quit()


    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def Next(self):
        self.parent.nx(None)

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def Previous(self):
        self.parent.nx(None, prev=True)

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def Pause(self):
        self.parent.pause()

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def PlayPause(self):
        self.parent.play_pause(None)

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def Stop(self):
        self.parent.pause()

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='', out_signature='')
    def Play(self):
        self.parent.play()

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='x', out_signature='')
    def Seek(self, seek_by):
        self.parent.seek_by(seek_by/1000000)

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='ox', out_signature='')
    def SetPosition(self, track_id, seek_to):
        # TODO: Check that track_id matches the currently playing track
        self.parent.seektime(None, seek_to)

    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player',
                         in_signature='s', out_signature='')
    def OpenUri(self, f):
        self.parent.add_file(None, str(f))

    @dbus.service.signal(dbus_interface='org.mpris.MediaPlayer2.Player',
                         signature='x')
    def Seeked(self, pos):
        pass
