#!/usr/bin/env python2
#
#  Valosoitin - Audio player for EFL
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import sys
import os
import time
import urllib
import argparse
import datetime
import cgi
import ConfigParser
import cPickle
import random
random.seed()
import mimetypes
mimetypes.init()
import logging
logging.basicConfig(level=logging.DEBUG)
import gettext
gettext.install("valosoitin", unicode=True)

import mutagen
import mutagen.mp3
import mutagen.id3
import mutagen.flac
import mutagen.oggvorbis

import elementary as elm
from ecore import Timer, Idler
from emotion import Emotion

from ValoDBus import ValosoitinDBus
from gui import PlayControls, PlayModes, VolumeControls, PlayList, PlayListControls, SeekControls, AudioInfo

try:
    from gi.repository import Notify
except ImportError:
    def notify(*args):
        pass
else:
    Notify.init("Valosoitin")
    def notify(title, msg, importance):
        n = Notify.Notification.new(title, msg, importance)
        n.set_timeout(3000)
        n.show()

parser = argparse.ArgumentParser(description='Utters sounds.')
parser.add_argument("files", metavar="file", type=str, nargs="*",
                    help="Files to add to the playlist")
clargs = parser.parse_args(sys.argv[1:])

import dbus, e_dbus
ml = e_dbus.DBusEcoreMainLoop()
dbus.set_default_main_loop(ml)
import dbus.service
bus = dbus.SessionBus()

dbo = None
try:
    dbo = bus.get_object("org.mpris.MediaPlayer2.Valosoitin", "/org/mpris/MediaPlayer2")
except dbus.exceptions.DBusException:
    pass

if dbo:
    if clargs.files:
        for f in clargs.files:
            dbo.OpenUri(f, dbus_interface="org.mpris.MediaPlayer2.Player")
    sys.exit()

conf = ConfigParser.SafeConfigParser()
conf_dir = os.path.expanduser(os.path.join("~", ".config", "valosoitin"))
conf_path = os.path.join(conf_dir, "valosoitin.conf")
if not os.path.exists(conf_dir):
    os.mkdir(conf_dir, 0700)
if not os.path.exists(os.path.join(conf_path)):
    conf.add_section("Settings")
    conf.set("Settings", "audio_backend", "gstreamer")
    conf.set("Settings", "single", "False")
    conf.set("Settings", "random", "False")
    conf.set("Settings", "repeat", "False")
    conf.set("Settings", "volume", "100.0")
    conf.set("Settings", "mute", "False")
    conf.add_section("Last played")
    conf.set("Last played", "filename", "")
    with open(conf_path, 'wb') as configfile:
        conf.write(configfile)
conf.read(conf_path)
for section in ("Settings", "Last played"):
    if not conf.has_section(section):
        conf.add_section(section)

try:
    if not conf.get("Settings", "audio_backend") in ("gstreamer", "xine"):
        logging.error("Invalid audio backend in config file. Set it to either gstreamer or xine.")
        sys.exit()
except:
    logging.error("audio_backend was not found in settings file. Set it to a value or remove the settings file so that a new one will be generated at next start of this program.")

class Valosoitin(object):
    def __init__(self, *args):
        self.conf = conf

        self.duration = datetime.timedelta()
        self.position = datetime.timedelta()

        self.seeking = False
        self.playtimer = None

        self.library = {}
        self.randomorder = []

        self.progress_bar_notify = None

        window = self.window = elm.StandardWindow("valosoitin", "Valosoitin")
        window.callback_delete_request_add(self.quit)
        window.callback_withdrawn_add(lambda x: self.playtimer.freeze())
        window.callback_iconified_add(lambda x: self.playtimer.freeze())
        window.callback_normal_add(self.normal_window)

        # GUI
        self.playcontrols = PlayControls(self)
        self.volumecontrols = VolumeControls(self)
        self.playmodes = PlayModes(self)
        self.seekcontrols = SeekControls(self)
        self.audioinfo = AudioInfo(self)
        self.playlist = PlayList(self)
        self.playlistcontrols = PlayListControls(self)

        # Top left box
        topleftbox = elm.Box(window)
        topleftbox.pack_end(self.playcontrols)
        topleftbox.pack_end(self.volumecontrols)
        topleftbox.pack_end(self.playmodes)
        topleftbox.show()

        # Top right box
        toprightbox = elm.Box(window)
        toprightbox.size_hint_weight_set(1.0, 0.0)
        toprightbox.size_hint_align_set(-1.0, 0.0)
        toprightbox.pack_end(self.seekcontrols)
        toprightbox.pack_end(self.audioinfo)
        toprightbox.show()

        # Top box
        topbox = elm.Box(window)
        topbox.horizontal_set(True)
        topbox.size_hint_weight_set(1.0, 0.0)
        topbox.size_hint_align_set(-1.0, 0.0)
        topbox.pack_end(topleftbox)
        topbox.pack_end(toprightbox)
        topbox.show()

        # Main box
        mainbox = elm.Box(window)
        mainbox.size_hint_weight_set(1.0, 1.0)
        window.resize_object_add(mainbox)
        mainbox.pack_end(topbox)
        mainbox.pack_end(self.playlist)
        mainbox.pack_end(self.playlistcontrols)
        mainbox.show()

        # Setup GUI update timer
        self.playtimer = playtimer = Timer(0.2, self.update)
        playtimer.freeze()

        # Setup Emotion
        self.player = player = Emotion(window.evas_get(), module_filename=conf.get("Settings", "audio_backend"))
        player.video_mute_set(True)
        player.callback_add("playback_started", self.playing)
        player.callback_add("playback_finished", self.finished)

        self.dbusname = dbus.service.BusName("org.mpris.MediaPlayer2.Valosoitin", dbus.SessionBus())
        self.dbo = ValosoitinDBus(self)

        window.resize(480, 480)
        window.show()

        # Restore playlist
        order = []
        try:
            with open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.library")), 'rb') as pkl_file:
                l = cPickle.load(pkl_file)
                if not isinstance(l, list):
                    self.library = l

            with open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.playlist_order")), 'rb') as pkl_file:
                order = cPickle.load(pkl_file)
        except:
            logging.exception("Could not restore playlist")

        paths = []
        for i in order:
            paths.append(self.library[i])

        for m in paths:
            f = m.filename
            if not os.path.exists(f):
                del self.library[f]

            self.playlist.add_item(self, m)

        f = conf.get("Last played", "filename")
        if f:
            if os.path.exists(f) and self.library.has_key(f):
                self.play_file(f, paused=True)

        self.playlist.show()
        self.randomize()

        self.change_volume(conf.getfloat("Settings", "volume"))
        self.mute(conf.getboolean("Settings", "mute"))

        # Add files from command line
        if clargs.files:
            self.add_files_with_idler(clargs.files)

    def raise_window(self):
        self.window.raise_()

    def normal_window(self, *args):
        if self.player.play_get():
            self.playtimer.thaw()
        else:
            self.update()

    def update(self):
        pos = self.player.position_get()
        self.dbo.playerprops["Position"] = pos*1000000

        if not self.seeking:
            self.seekcontrols.seek.value = pos
        pos = datetime.timedelta(seconds=int(pos))
        self.position = pos
        t = "%s / %s" % (pos, self.duration)
        self.seekcontrols.seektime.text_set(t)

        dur = self.player.play_length
        if dur != self.duration:
            self.seekcontrols.seek.min_max_set(0, dur)
            dur = datetime.timedelta(seconds=int(dur))
            self.duration = dur
            self.dbo.playerprops["Metadata"]["mpris:length"] = int(dur.total_seconds()) * 1000000

        return 1

    def add_files_with_idler(self, paths):
        if len(paths) > 200:
            for widget in self.playlistcontrols.children:
                widget.disabled = True
            self.audioinfo.search.disabled = True
            self.pbaru = 1.0/len(paths)
            pbarn = self.progress_bar_notify = elm.Notify(self.window)
            pbarn.orient = elm.ELM_NOTIFY_ORIENT_BOTTOM
            pbar = self.pbar = elm.Progressbar(self.window)
            pbarn.content = pbar
            pbar.span_size = self.window.size[0] - 100
            pbarn.show()

        logging.debug("Adding files")
        self.time1 = time.clock()
        logging.debug(self.time1)

        path_generator = (f for f in paths)
        idler = Idler(self.add_file, path_generator)

    def add_file(self, paths):
        try:
            f = paths.next()
        except StopIteration:
            self.time2 = time.clock()
            logging.debug(self.time2)
            logging.debug(self.time2 - self.time1)
            if self.progress_bar_notify:
                self.progress_bar_notify.delete()
                self.progress_bar_notify = None
                for widget in self.playlistcontrols.children:
                    widget.disabled = False
                self.audioinfo.search.disabled = False
                self.audioinfo.search.focus = True
            self.randomize()
            return False

        if self.progress_bar_notify:
            self.pbar.value = self.pbar.value + self.pbaru

        if not os.path.exists(f):
            logging.info("File doesn't exist")
            return True

        m = self.check_mime_type(f)
        if m is None:
            return True

        # Playlist items are unique by filepath
        if not self.library.has_key(f):
            self.library[f] = m

            self.playlist.add_item(self, m)
        else:
            logging.info("Audio already on playlist")

        return True

    def check_mime_type(self, f):
        mimetype = mimetypes.guess_type(f, strict=False)[0]
        if not mimetype:
            logging.info("Invalid type %s for file %s" % (mimetype, f))
            return None
        elif not mimetype.startswith("audio"):
            logging.info("Invalid type %s for file %s" % (mimetype, f))
            return None
        
        m = None

        if mimetype == "audio/mpeg":
            try:
                m = EasyID3(f)
            except mutagen.id3.ID3NoHeaderError:
                try:
                    m = mutagen.mp3.MP3(f)
                except mutagen.mp3.HeaderNotFoundError:
                    logging.info("Mpeg audio header not found, trying with generic type...")
                    m = mutagen.File(f, easy=True)
        elif mimetype == "audio/flac":
            try:
                m = mutagen.flac.FLAC(f)
            except mutagen.flac.FLACNoHeaderError:
                logging.info("FLAC metadata not found in file {0}.".format(f))
        elif mimetype == "audio/ogg":
            try:
                m = mutagen.oggvorbis.OggVorbis(f)
            except mutagen.oggvorbis.OggVorbisHeaderError:
                logging.info("Ogg metadata not found in file {0}.".format(f))
        else:
            logging.info("No known mutagen type for %s" % mimetype)
            m = mutagen.File(f, easy=True)

        #~ "audio/basic",
        #~ "audio/mid",
        #~ "audio/mp4",
        #~ "audio/ogg",
        #~ "audio/vorbis",
        #~ "audio/wav",
        #~ "audio/x-wav",

        return m

    def add_dir(self, path):
        if not path:
            return
        paths = []
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                paths.append(os.path.join(dirpath, filename))
        paths.sort()
        self.add_files_with_idler(paths)

    def play_file(self, f, paused=False):
        current_file = self.player.file_get()
        current_item = None
        new_item = None

        if current_file:
            try:
                current_item = self.playlist.items[urllib.unquote(current_file.split("://")[1])]
            except KeyError:
                pass

        try:
            new_item = self.playlist.items[f]
        except KeyError:
            logging.exception("Item for {0} not found".format(f))

        self.player.file_set("file://"+urllib.quote(f))
        if not paused:
            self.play()

        if current_item:
            current_item.fields_update("elm.swallow.icon", elm.ELM_GENLIST_ITEM_FIELD_CONTENT)
        if new_item:
            new_item.fields_update("elm.swallow.icon", elm.ELM_GENLIST_ITEM_FIELD_CONTENT)

    def play_pause(self, *args):
        if self.player.play_get():
            self.pause()
        else:
            self.play()

    def play(self):
        if self.player.play_get():
            return
        ic = elm.Icon(self.window)
        if not self.player.file_get():
            i = self.playlist.selected_item
            if not i:
                try:
                    f = self.playlist.order[0]
                except:
                    logging.info("No files on the playlist")
                    return
                i = self.playlist.items[f]
            m = i.data_get()
            self.play_file(m.filename)
            return
        self.player.play_set(True)
        ic.standard_set("media-playback-pause")
        self.playcontrols.play.content_set(ic)
        self.playtimer.thaw()
        self.dbo.playerprops["PlaybackStatus"] = "Playing"

    def pause(self, override=False):
        if not self.player.play_get() and not override:
            return
        ic = elm.Icon(self.window)
        self.player.play_set(False)
        ic.standard_set("media-playback-start")
        self.playcontrols.play.content_set(ic)
        self.playcontrols.play.size_hint_min_set(30, 30)
        self.playtimer.freeze()
        self.dbo.playerprops["PlaybackStatus"] = "Paused"

    def nx(self, btn, prev=False):
        f = self.player.file_get()
        if f:
            f = urllib.unquote(f).split("://")[1]
        else:
            return
        c_i = self.playlist.order.index(f)
        if self.playmodes.random.state:
            i = self.randomorder.index(c_i)
            if prev:
                i = i-1
                if i < 0:
                    if self.playmodes.repeat.state:
                        i = len(self.playlist.order)-1
                    else:
                        return
            else:
                i = i+1
                if i > len(self.playlist.order)-1:
                    if self.playmodes.repeat.state:
                        i = 0
                    else:
                        return
            next_i = self.randomorder[i]
        else:
            if prev:
                next_i = c_i - 1
                if next_i < 0:
                    next_i = None
            else:
                next_i = c_i + 1
                if next_i >= len(self.playlist.order):
                    next_i = None
        if not next_i:
            if self.playmodes.repeat.state:
                next_i = 0
            else:
                if not btn:
                    # End of playlist
                    self.pause(True)
                return
        f = self.playlist.order[next_i]
        self.play_file(f)

    def seekstart(self, slider):
        self.seeking = True

    def seektime(self, slider, t=None):
        self.seeking = False
        t = t if t is not None else slider.value
        ft = datetime.timedelta(seconds=int(t))
        d = self.duration
        self.seekcontrols.seektime.text_set("%s / %s" % (ft, d))
        self.player.position_set(t)
        self.dbo.Seeked(t*1000000)

    def seek_by(self, t):
        ct = self.player.position_get()
        t = ct+t
        self.seektime(None, t)

    def change_volume(self, value):
        self.player.audio_volume_set(value/100)
        self.dbo.playerprops["Volume"] = value/100
        self.conf.set("Settings", "volume", str(value))

    def mute(self, state):
        self.volumecontrols.volume.disabled = state
        self.player.audio_mute_set(state)
        self.conf.set("Settings", "mute", str(state))

    def set_single_mode(self, mode):
        if self.playmodes.repeat.state:
            if mode:
                v = "Track"
            else:
                v = "Playlist"
        else:
            v = "None"
        self.dbo.playerprops["LoopStatus"] = mode
        self.conf.set("Settings", "single", str(mode))

    def set_repeat_mode(self, mode):
        if mode:
            if self.playmodes.single.state:
                v = "Track"
            else:
                v = "Playlist"
        else:
            v = "None"
        self.dbo.playerprops["LoopStatus"] = mode
        self.conf.set("Settings", "repeat", str(mode))

    def set_random_mode(self, mode):
        self.dbo.playerprops["Shuffle"] = mode
        self.conf.set("Settings", "random", str(mode))

    def playing(self, player):
        f = urllib.unquote(self.player.file_get()).split("://")[1]
        m = self.library[f]

        titles = cgi.escape(", ".join(m.get("title"))) if m.get("title") else os.path.splitext(os.path.basename(m.filename))[0]
        artists = cgi.escape(", ".join(m.get("artist"))) if m.get("artist") else ""
        albums = cgi.escape(", ".join(m.get("album"))) if m.get("album") else ""

        self.audioinfo.title.text = _("title <b>{0}</b>").format(titles)
        self.audioinfo.artist.text = _("by <i>{0}</i>").format(artists) if artists else ""
        self.audioinfo.album.text = _("from the album <b>{0}</b>").format(albums) if albums else ""

        self.dbo.playerprops["Metadata"]["mpris:trackid"] = f

        self.dbo.playerprops["Metadata"]["xesam:title"] = titles
        self.dbo.playerprops["Metadata"]["xesam:artist"] = artists
        self.dbo.playerprops["Metadata"]["xesam:album"] = albums

        notify(_("You're listening to"), "\n".join(filter(None, (artists, titles, albums))), "notification-information")

    def finished(self, player):
        if self.playmodes.single.state:
            if self.playmodes.repeat.state:
                self.seektime(None, 0)
                self.play()
            else:
                self.pause(True)
        else:
            self.nx(None)

    def randomize(self):
        # Create a list from 0 to length of playlist, then shuffle it
        o = [x for x in range(self.playlist.items_count())]
        random.shuffle(o)
        self.randomorder = o

    def search(self, string, cb):
        logging.debug("Searching")
        time1 = time.clock()
        logging.debug(time1)

        result = {}
        for k, m in self.library.iteritems():
            if not m:
                if string.lower() in k.decode("utf-8").lower():
                    result[k] = m
                continue

            for tag in m.itervalues():
                for v in tag:
                    if string.lower() in v.lower():
                        result[k] = m

        ordered_result = []
        for k in sorted(result.iterkeys()):
            ordered_result.append(result[k])

        time2 = time.clock()
        logging.debug(time2)
        logging.debug(time2 - time1)

        if result:
            cb(string, ordered_result)
        else:
            n = elm.Notify(self.window)
            n.orient = elm.ELM_NOTIFY_ORIENT_TOP
            l = elm.Label(self.window)
            l.text = _("No results for {0}").format(string)
            n.content = l
            n.timeout = 2.0
            n.show()

    def quit(self, *args):

        try:
            with open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.library")), 'wb') as output:
                cPickle.dump(self.library, output, 2)
        except:
            logging.exception("Could not save library")

        try:
            with open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.playlist_order")), 'wb') as output:
                cPickle.dump(self.playlist.order, output, 2)
        except:
            logging.exception("Could not save playlist order")

        try:
            conf.set("Last played", "filename", urllib.unquote(self.player.file_get()).split("://")[1] if self.player.file_get() else "")
            with open(conf_path, 'wb') as configfile:
                conf.write(configfile)
        except:
            logging.exception("Could not save conf")

        elm.exit()

# XXX: Monkey patching mutagen to fix http://code.google.com/p/mutagen/issues/detail?id=102
#      This can be removed once 1.21 becomes common
class EasyID3(mutagen.easyid3.EasyID3):
    def __init__(self, filename=None):
        self.__id3 = mutagen.id3.ID3()
        if filename is not None:
            self.load(filename)

    load = property(lambda s: s.__id3.load,
                    lambda s, v: setattr(s.__id3, 'load', v))

    save = property(lambda s: s.__id3.save,
                    lambda s, v: setattr(s.__id3, 'save', v))

    delete = property(lambda s: s.__id3.delete,
                      lambda s, v: setattr(s.__id3, 'delete', v))

if __name__ == "__main__":
    elm.init()
    valosoitin = Valosoitin(None)
    elm.run()
    elm.shutdown()
