#!/usr/bin/env python2
#
#  Valosoitin - Audio player for EFL
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import os
import cgi
import urllib
import logging

from evas import EVAS_ASPECT_CONTROL_VERTICAL
import elementary as elm

class PlayList(elm.Genlist):
    def __init__(self, parent):
        self.vs = parent
        window = parent.window
        elm.Genlist.__init__(self, window)

        self.items = {}
        self.order = []

        self.homogeneous = True
        self.multi_select_set(False)
        self.reorder_mode_set(True)
        self.callback_selected_add(self.item_selected, parent)
        self.callback_activated_add(self.play_selected, parent)
        self.callback_moved_add(self.item_moved, parent)
        self.size_hint_weight_set(1.0, 1.0)
        self.size_hint_align_set(-1.0, -1.0)
        #self.show()

    def item_selected(self, genlist, item, parent):
        i = item.index_get()
        l = genlist.items_count()
        parent.playlistcontrols.plinfo.text_set("%s / %s" % (i, l))

    def play_selected(self, genlist, item, parent):
        m = item.data_get()
        parent.play_file(m.filename)

    def item_moved(self, gl, item, parent):
        i = parent.playlist.order
        o = i.index(item.data.filename)
        if item.prev:
            p = i.index(item.prev.data.filename)
            if o > p:
                p = p+1
            i.insert(p, i.pop(o))
        else:
            i.insert(0, i.pop(o))

    def add_item(self, vs, m):
        item = self.item_append(PlayListItemClass(), m, None, elm.ELM_GENLIST_ITEM_NONE)
        f = m.filename
        self.items[f] = item
        self.order.append(f)

class PlayListControls(elm.Box):
    def __init__(self, parent):
        window = parent.window
        elm.Box.__init__(self, window)

        addFile = elm.FileselectorButton(window)
        addFile.text = _("Add file")
        addFile.inwin_mode = True
        addFile.is_save = False
        addFile.callback_file_chosen_add(self.add_file_cb, parent)
        addFile.show()

        addDir = elm.FileselectorButton(window)
        addDir.text = _("Add directory")
        addDir.folder_only = True
        addDir.inwin_mode = True
        addDir.is_save = False
        addDir.expandable = False
        addDir.callback_file_chosen_add(self.add_dir_cb, parent)
        addDir.show()

        rm = elm.Button(window)
        rm.text = _("Remove")
        rm.callback_clicked_add(self.remove_selected_cb, parent)
        rm.show()

        cl = elm.Button(window)
        cl.text = _("Clear")
        cl.tooltip_text_set(_("Clears your entire playlist"))
        cl.callback_clicked_add(self.clear_playlist, parent)
        cl.show()

        bi = elm.Button(window)
        bi.text = _("Bring in")
        bi.tooltip_text_set(_("Scrolls the playlist to display your currently playing track"))
        bi.callback_clicked_add(lambda x: parent.playlist.items[urllib.unquote(parent.player.file_get()).split("://")[1]].bring_in() if parent.player.file_get() else parent.player.file_get())
        bi.show()

        self.plinfo = plinfo = elm.Label(window)
        plinfo.size_hint_weight_set(1.0, 0.0)
        plinfo.size_hint_align_set(1.0, 0.5)
        plinfo.text_set("")
        plinfo.show()

        self.horizontal_set(True)
        self.size_hint_weight_set(1.0, 0.0)
        self.size_hint_align_set(-1.0, 0.0)
        self.pack_end(addFile)
        self.pack_end(addDir)
        self.pack_end(rm)
        self.pack_end(cl)
        self.pack_end(bi)
        self.pack_end(plinfo)
        self.show()

    def add_file_cb(self, fs, path, parent):
        if not path:
            logging.info("No file selected")
            return
        parent.add_files_with_idler((path,))

    def add_dir_cb(self, fs, path, parent):
        if not path:
            logging.info("No dir selected")
            return
        parent.add_dir(path)

    def remove_selected_cb(self, btn, parent):
        if parent.playlist.multi_select_get():
            self.n.delete()
            del self.n
            btn.style = "default"
            parent.playlist.multi_select_set(False)
            self.remove_selected(parent)
        else:
            btn.style = "anchor"
            i = parent.playlist.selected_item
            if i:
                i.selected_set(False)
            parent.playlist.multi_select_set(True)
            l = elm.Label(parent.window)
            l.text = _("Select the tracks you wish to remove and press Remove when done.")
            n = self.n = elm.Notify(parent.window)
            n.orient = elm.ELM_NOTIFY_ORIENT_TOP
            n.content = l
            n.show()

    def remove_selected(self, parent):
        items = parent.playlist.selected_items_get()
        if not items:
            return
        for item in items:
            m = item.data_get()
            item.delete()
            f = m.filename
            del parent.playlist.items[f]
            parent.playlist.order.remove(f)
            del parent.library[f]

    def clear_playlist(self, btn, parent):
        parent.playlist.clear()
        parent.playlist.items.clear()
        del parent.playlist.order[:]
        parent.library.clear()

class PlayListItemClass(elm.GenlistItemClass):
    def text_get(self, genlist, part, info):
        m = info

        artists = ", ".join(m.get("artist")) if m.get("artist") else ""
        albums = ", ".join(m.get("album")) if m.get("album") else ""
        titles = ", ".join(m.get("title")) if m.get("title") else os.path.basename(m.filename)

        strings = (artists, albums, titles)
        entry = cgi.escape(" - ".join(filter(None, strings)))
        return entry

    def content_get(self, genlist, part, info):
        m = info
        player = genlist.vs.player

        if part == "elm.swallow.icon":
            ic = elm.Icon(genlist)
            if player.file_get() == "file://"+urllib.quote(m.filename):
                ic.standard_set("media-playback-start")
            else:
                ic.standard_set("media-playback-pause")
            ic.size_hint_aspect_set(EVAS_ASPECT_CONTROL_VERTICAL, 1, 1)
            return ic
